"""
A module that contains utility functions to load the 'classical' workspace configuration.
This configuration may have three meaningful files:
.remote (required) - information about the connection options
.remoteindex (optional) - information about which connection from options above to use
.remoteignore (optional) - information about files that should be ignore when syncing files
"""
import hashlib
import os
import re

from collections import defaultdict
from dataclasses import asdict
from pathlib import Path
from typing import Dict, List, Tuple

from remote.exceptions import ConfigurationError

from . import ConfigurationMedium, RemoteConfig, SyncIgnores, WorkspaceConfig

CONFIG_FILE_NAME = ".remote"
INDEX_FILE_NAME = ".remoteindex"
IGNORE_FILE_NAME = ".remoteignore"
IGNORE_SECTION_REGEX = re.compile(r"^(push|pull|both)\s*:$")
BASE_IGNORES = (CONFIG_FILE_NAME, INDEX_FILE_NAME, IGNORE_FILE_NAME)
DEFAULT_SHELL = "sh"
DEFAULT_SHELL_OPTIONS = ""


def _extract_shell_info(line: str, env_vars: List[str]) -> Tuple[str, str]:
    if not env_vars:
        return DEFAULT_SHELL, DEFAULT_SHELL_OPTIONS

    vars_string = env_vars[0]

    env = {}
    items = vars_string.split()
    index = 0
    while index < len(items):
        key, value = items[index].split("=")
        if value.startswith("'") or value.startswith('"'):
            control_character = value[0]
            while index < len(items) - 1:
                if value[-1] == control_character:
                    break
                index += 1
                value += " " + items[index]
            if not value[-1] == control_character:
                raise ConfigurationError(f"Config line {line} is corrupted. Cannot parse end {key}={value}")

        env[key] = value.strip("\"'")

        index += 1
    print(env)
    # TODO: these shell types are not used in new implementation, need to remove them
    shell = env.pop("RSHELL", DEFAULT_SHELL)
    shell_options = env.pop("RSHELL_OPTS", DEFAULT_SHELL_OPTIONS)
    if env:
        raise ConfigurationError(
            f"Config line {line} contains unexpected env variables: {env}. Only RSHELL and RSHELL_OPTS can be used"
        )
    return shell, shell_options


def parse_config_line(line: str) -> RemoteConfig:
    # The line should look like this:
    # sdas-ld2:.remotes/814f27f15f4e7a0842cada353dfc765a RSHELL=zsh
    entry, *env_items = line.split(maxsplit=1)
    shell, shell_options = _extract_shell_info(line, env_items)

    parts = entry.split(":")
    if len(parts) != 2:
        raise ConfigurationError(
            f"The configuration string is malformed: {parts}. Please use host-name:remote_dir format"
        )
    host, directory = parts
    return RemoteConfig(host=host, directory=Path(directory), shell=shell, shell_options=shell_options)


def load_configurations(workspace_root: Path) -> List[RemoteConfig]:
    config_file = workspace_root / CONFIG_FILE_NAME

    configurations = []
    for line in config_file.read_text().splitlines():
        line = line.strip()
        if not line or line.startswith("#"):
            continue

        configurations.append(parse_config_line(line))

    return configurations


def load_default_configuration_num(workspace_root: Path) -> int:
    # If REMOTE_HOST_INDEX is set, that overrides settings in .remoteindex
    env_index = os.environ.get("REMOTE_HOST_INDEX")
    if env_index:
        try:
            return int(env_index)
        except ValueError:
            raise ConfigurationError(
                f"REMOTE_HOST_INDEX env variable contains symbols other than numbers: '{env_index}'. "
                "Please set the coorect index value to continue"
            )

    index_file = workspace_root / INDEX_FILE_NAME
    if not index_file.exists():
        return 0

    # Configuration uses 1-base index and we need to have 0-based
    text = index_file.read_text().strip()
    try:
        return int(text) - 1
    except ValueError:
        raise ConfigurationError(
            f"File {index_file} contains symbols other than numbers: '{text}'. "
            "Please remove it or replace the value to continue"
        )


def _postprocess(ignores):
    pull = ignores.pop("pull", [])
    push = ignores.pop("push", [])
    both = ignores.pop("both", [])
    if ignores:
        raise ConfigurationError(
            f"{IGNORE_FILE_NAME} file has unexpected sections: {', '.join(ignores.keys())}. Please remove them"
        )
    return SyncIgnores(pull=pull, push=push, both=both)


def load_ignores(workspace_root: Path) -> SyncIgnores:
    ignores: Dict[str, List[str]] = defaultdict(list)
    ignores["both"].extend(BASE_IGNORES)

    ignore_file = workspace_root / IGNORE_FILE_NAME
    if not ignore_file.exists():
        return _postprocess(ignores)

    active_section = "both"
    is_new_format = None
    for line in ignore_file.read_text().splitlines():
        line = line.strip()
        if not line or line.startswith("#"):
            continue

        matcher = IGNORE_SECTION_REGEX.match(line)
        if matcher is None:
            if is_new_format is None:
                is_new_format = False
            ignores[active_section].append(line)
        else:
            if is_new_format is None:
                is_new_format = True
            elif not is_new_format:
                raise ConfigurationError(
                    f"Few ignore patters were listed in {IGNORE_FILE_NAME} before the first section {matcher.group(1)} appeared. "
                    "Please list all ignored files after a section declaration if you use new ignore format"
                )
            active_section = matcher.group(1)

    return _postprocess(ignores)


def save_general_config(config_file: Path, configurations: List[RemoteConfig]):
    with config_file.open("w") as f:
        for item in configurations:
            f.write(f"{item.host}:{item.directory}")
            if item.shell != "sh":
                f.write(f" RSHELL={item.shell}")
            if item.shell_options:
                f.write(f" RSHELL_OPTS='{item.shell_options}'")
            f.write("\n")


def save_ignores(config_file: Path, ignores: SyncIgnores):
    ignores.both.extend(BASE_IGNORES)
    ignores.trim()

    if ignores.is_empty():
        if config_file.exists():
            config_file.unlink()
        return

    with config_file.open("w") as f:
        for key, value in asdict(ignores).items():
            f.write(f"{key}:\n")
            for item in value:
                f.write(f"{item}\n")


def save_index(config_file: Path, index: int):
    if index == 0:
        # We delete file when index is default
        if config_file.exists():
            config_file.unlink()
    else:
        config_file.write_text(f"{index + 1}\n")


class ClassicConfigurationMedium(ConfigurationMedium):
    """A medium class that knows hot to load and save the 'classical' workspace configuration.

    This configuration may have three meaningful files:
    .remote (required) - information about the connection options
    .remoteindex (optional) - information about which connection from options above to use
    .remoteignore (optional) - information about files that should be ignore when syncing files
    """

    def load_config(self, workspace_root: Path) -> WorkspaceConfig:
        configurations = load_configurations(workspace_root)
        configuration_index = load_default_configuration_num(workspace_root)
        if configuration_index > len(configurations) - 1:
            raise ConfigurationError(
                f"Configuration #{configuration_index + 1} requested but there are only {len(configurations)} declared"
            )
        ignores = load_ignores(workspace_root)

        return WorkspaceConfig(
            root=workspace_root,
            configurations=configurations,
            default_configuration=configuration_index,
            ignores=ignores,
        )

    def save_config(self, config: WorkspaceConfig):
        save_general_config(config.root / CONFIG_FILE_NAME, config.configurations)
        save_ignores(config.root / IGNORE_FILE_NAME, config.ignores)
        save_index(config.root / INDEX_FILE_NAME, config.default_configuration)

    def is_workspace_root(self, path: Path) -> bool:
        return (path / CONFIG_FILE_NAME).exists()

    def generate_remote_directory(self, config: WorkspaceConfig) -> Path:
        md5 = hashlib.md5(str(config.root).encode()).hexdigest()
        return Path(f".remotes/{config.root.name}_{md5[:8]}")



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Tuple


@dataclass
class RemoteConfig:
    """Single remote connection description"""

    # remote machine's hostname
    host: str
    # relative path to the working directory on remote machine starting from user home dir
    directory: Path
    # a shell to use on remote machine
    shell: str
    # shell options to use on remote machine
    shell_options: str


@dataclass
class SyncIgnores:
    """Patterns used to ignore files when syncing with remote location"""

    # patterns to ignore while pulling from remote
    pull: List[str]
    # patterns to ignore while pushing from local
    push: List[str]
    # patterns to ignore while transferring files in both directions
    both: List[str]

    def __post_init__(self):
        self.trim()

    def compile_push_ignores(self):
        result = set()
        result.update(self.push)
        result.update(self.both)
        return sorted(result)

    def compile_pull_ignores(self):
        result = set()
        result.update(self.pull)
        result.update(self.both)
        return sorted(result)

    def add_ignores(self, ignores: List[str]):
        new_ignores = set()
        new_ignores.update(ignores)
        new_ignores.update(self.both)
        self.both = sorted(new_ignores)

    def trim(self):
        self.pull = sorted(set(self.pull))
        self.push = sorted(set(self.push))
        self.both = sorted(set(self.both))

    def is_empty(self):
        return not (self.pull or self.push or self.both)

    @classmethod
    def new(cls) -> "SyncIgnores":
        return cls([], [], [])


@dataclass
class WorkspaceConfig:
    """Complete remote workspace config"""

    # absolute path to the workspace root
    root: Path
    # remote host connection options that can be used in this workspace
    configurations: List[RemoteConfig]
    # index of default remote host connection
    default_configuration: int
    # patterns to ignore while syncing the workspace
    ignores: SyncIgnores

    @classmethod
    def empty(cls, root: Path) -> "WorkspaceConfig":
        return cls(root=root, configurations=[], default_configuration=0, ignores=SyncIgnores.new())

    def add_remote_host(
        self, host: str, directory: Path, shell: Optional[str] = None, shell_options: Optional[str] = None
    ) -> Tuple[bool, int]:
        remote_config = RemoteConfig(
            host=host, directory=directory, shell=shell or "sh", shell_options=shell_options or "",
        )
        for num, cfg in enumerate(self.configurations):
            if cfg.host == remote_config.host and cfg.directory == remote_config.directory:
                return False, num
        self.configurations.append(remote_config)
        return True, len(self.configurations) - 1


class ConfigurationMedium(metaclass=ABCMeta):
    """A medium class that knows how to load, save, or process a certain type of configuration layout"""

    @abstractmethod
    def load_config(self, workspace_root: Path) -> WorkspaceConfig:
        """Load configuration for the workspace that is located in provided root directory.
        If this method is called, we could assume that check in `is_workspace_root` passed
        """

    @abstractmethod
    def save_config(self, config: WorkspaceConfig):
        """Save configuration to its root"""

    @abstractmethod
    def is_workspace_root(self, path: Path) -> bool:
        """Return true is the path provided contains a configured workspace that can be loaded by this medium"""

    @abstractmethod
    def generate_remote_directory(self, config: WorkspaceConfig) -> Path:
        """Renerate a default remote directory path for the workspace with provided configuration"""



import logging
import re
import subprocess
import sys
import tempfile
import time

from contextlib import contextmanager
from pathlib import Path
from typing import List, Sequence, Union

from .exceptions import RemoteConnectionError, RemoteExecutionError

logger = logging.getLogger(__name__)


def _temp_file(lines: List[str]) -> Path:
    """Create a temporary file with provided content and return its path

    :param lines: list of lines to be written in the file
    """
    _, path = tempfile.mkstemp(prefix="remote.", dir="/tmp", text=True)
    tmpfile = Path(path)
    tmpfile.write_text("\n".join(lines) + "\n")

    return tmpfile


def _gen_rsync_patterns_file(patterns, opt, args, cleanup):
    if patterns:
        exclude_file = _temp_file(patterns)
        cleanup.append(exclude_file)
        args.extend((opt, str(exclude_file)))
        logger.info(f"{opt} patterns:")
        for p in patterns:
            logger.info("  - %s", p)


@contextmanager
def _measure_duration(operation: str):
    start = time.time()
    yield None
    runtime = time.time() - start
    logger.info("%s done in %.2f seconds", operation, runtime)


def rsync(
    src: str,
    dst: str,
    info: bool = False,
    verbose: bool = False,
    dry_run: bool = False,
    mirror: bool = False,
    excludes: List[str] = None,
    includes: List[str] = None,
    extra_args: List[str] = None,
):
    """Run rsync to sync files from src into dst

    :param src: Source files to copy. If source is a directory and you need to copy its contents, append / to its path
    :param dst: Destination file or directory
    :param info: True if need to add -i flag to rsync
    :param verbose: True if need to add -v flag to rsync
    :param dry_run: True if need to add -n flag to rsync
    :param mirror: True if all files inside destination directory need to be deleted if they were not bringed from source
    :param excludes: List of file patterns to exclude from syncing
    """

    logger.info("Sync files from %s to %s", src, dst)
    args = ["rsync", "-rlpmchz", "--copy-unsafe-links", "-e", "ssh -q", "--force"]
    if info:
        args.append("-i")
    if verbose:
        args.append("-v")
    if dry_run:
        args.append("-n")
    if mirror:
        args.extend(("--delete", "--delete-after", "--delete-excluded"))
    if extra_args:
        args.extend(extra_args)

    cleanup: List[Path] = []
    # It is important to add include patterns before exclude patters because rsync might ignore includes if you do otherwise.
    _gen_rsync_patterns_file(includes, "--include-from", args, cleanup)
    _gen_rsync_patterns_file(excludes, "--exclude-from", args, cleanup)

    args.extend((src, dst))

    logger.info("Starting sync with command %s", " ".join(args))
    with _measure_duration("Sync"):
        result = subprocess.run(args, stdout=sys.stdout, stderr=sys.stderr)

    for file in cleanup:
        file.unlink()

    if result.returncode != 0:
        raise RemoteConnectionError(f"Failed to sync files betwen {src} and {dst}. Is remote host reachable")


def prepare_shell_command(command: Union[str, Sequence[str]]) -> str:
    """Format command parts into one shell command"""
    if isinstance(command, str):
        return command
    # This means the whole command is already preformatted for us
    if len(command) == 1 and " " in command[0]:
        return command[0]

    result = []
    for item in command:
        if not item:
            continue
        if re.search(r"\s+", item):
            result.append(f"'{item}'")
        else:
            result.append(item)

    return " ".join(result)


def ssh(host: str, command: str, dry_run: bool = False, raise_on_error: bool = True):
    """Execute a command remotely using SSH and return it's exit code

    :param host: a name of the host that will be used for execution
    :param command: a command to execute
    :param dry_run: log command instead of executing it
    :param raise_on_error: raise an exception is remote execution

    :returns: exit code of remote command or 255 if connection didn't go through
    """
    subprocess_command = ["ssh", "-tKq", host, command]
    logger.info("Executing:\n%s %s %s <<EOS\n%sEOS", *subprocess_command)
    if dry_run:
        return 0

    with _measure_duration("Execution"):
        result = subprocess.run(subprocess_command, stdout=sys.stdout, stderr=sys.stderr, stdin=sys.stdin)

    if raise_on_error:
        # ssh exits with the exit status of the remote command or with 255 if an error occurred
        if result.returncode == 255:
            raise RemoteConnectionError(f"Failed to connect to {host}")
        elif result.returncode != 0:
            raise RemoteExecutionError(f'Failed to execute "{command}" on host {host} ({result.returncode})')
    return result.returncode



"""Integration-ish tests for CLI endpoints.
We only mock subprocess calls since we cannot do multi-host testing.

Some of the test above don't verify much, but they at least ensure that all parts work well together.
"""
import os

from contextlib import contextmanager
from unittest.mock import ANY, Mock, call, patch

import pytest

from click import BadParameter
from click.testing import CliRunner

from remote import entrypoints
from remote.configuration.classic import CONFIG_FILE_NAME, IGNORE_FILE_NAME, INDEX_FILE_NAME
from remote.exceptions import RemoteExecutionError

TEST_HOST = "test-host1.example.com"
TEST_DIR = ".remotes/myproject"
TEST_CONFIG = f"{TEST_HOST}:{TEST_DIR}"


@contextmanager
def cwd(path):
    old_cwd = os.getcwd()
    os.chdir(str(path))
    try:
        yield path
    finally:
        os.chdir(old_cwd)


@pytest.fixture
def tmp_workspace(tmp_path):
    (tmp_path / CONFIG_FILE_NAME).write_text(TEST_CONFIG + "\n")
    (tmp_path / IGNORE_FILE_NAME).write_text(
        """\
pull:
push:
both:
.remote
.remoteignore
.remoteindex
"""
    )
    return tmp_path


def test_log_exceptions_decorator():
    @entrypoints.log_exceptions
    def test_function(num):
        if num > 0:
            raise RemoteExecutionError("Some execution error")
        elif num == 0:
            raise ValueError("Some value error")
        else:
            return

    # This shouldnt fail
    test_function(-1)

    # This should fail with custom error and cause sys.exit
    with pytest.raises(SystemExit):
        test_function(1)

    # This should fail with internal error and propagate exception
    with pytest.raises(ValueError):
        test_function(0)


@pytest.mark.parametrize(
    "connection, is_valid",
    [
        ("host", True),
        ("host123", True),
        ("host.domain.com", True),
        ("ho-st.dom-ain.as1234", True),
        ("ho-st.dom-ain.as1234:/home/dir", True),
        ("ho-st.dom-ain.as1234:.home/dir.dir", True),
        ("ho-st.dom-ain.as1234:.home/dir.dir/123/", True),
        ("ho-st.dom-ain.as1234:.home/dir.dir/123/:something", False),
        ("ho-st.dom-ain.as1234::/home/dir", False),
    ],
)
def test_validate_connection_string(connection, is_valid):
    if is_valid:
        entrypoints.validate_connection_string(None, None, connection)
    else:
        with pytest.raises(BadParameter):
            entrypoints.validate_connection_string(None, None, connection)


@patch("remote.util.subprocess.run")
def test_remote_init(mock_run, tmp_path):
    mock_run.return_value = Mock(returncode=0)
    subdir = tmp_path / "myproject"
    subdir.mkdir()

    runner = CliRunner()
    with cwd(subdir):
        result = runner.invoke(entrypoints.remote_init, ["test-host.example.com"])

    assert result.exit_code == 0
    assert "Created remote directory at test-host.example.com:.remotes/myproject_" in result.output
    assert "Remote is configured and ready to use" in result.output

    mock_run.assert_called_once_with(["ssh", "-tKq", "test-host.example.com", ANY], stdin=ANY, stdout=ANY, stderr=ANY)

    assert (subdir / CONFIG_FILE_NAME).exists()
    assert (subdir / CONFIG_FILE_NAME).read_text().startswith("test-host.example.com:.remotes/myproject_")
    assert (subdir / IGNORE_FILE_NAME).exists()
    assert (
        (subdir / IGNORE_FILE_NAME).read_text()
        == """\
pull:
push:
both:
.remote
.remoteignore
.remoteindex
"""
    )


@patch("remote.util.subprocess.run")
def test_remote_init_with_dir(mock_run, tmp_path):
    mock_run.return_value = Mock(returncode=0)
    subdir = tmp_path / "myproject"
    subdir.mkdir()

    runner = CliRunner()
    with cwd(subdir):
        result = runner.invoke(entrypoints.remote_init, ["test-host.example.com:.path/test.dir/_test-dir/"])

    assert result.exit_code == 0
    assert (
        result.output
        == """\
Created remote directory at test-host.example.com:.path/test.dir/_test-dir
Remote is configured and ready to use
"""
    )

    mock_run.assert_called_once_with(
        ["ssh", "-tKq", "test-host.example.com", "mkdir -p .path/test.dir/_test-dir"],
        stdin=ANY,
        stdout=ANY,
        stderr=ANY,
    )

    assert (subdir / CONFIG_FILE_NAME).exists()
    assert (subdir / CONFIG_FILE_NAME).read_text() == "test-host.example.com:.path/test.dir/_test-dir\n"
    assert (subdir / IGNORE_FILE_NAME).exists()
    assert (
        (subdir / IGNORE_FILE_NAME).read_text()
        == """\
pull:
push:
both:
.remote
.remoteignore
.remoteindex
"""
    )


@patch("remote.util.subprocess.run")
def test_remote_init_fails_after_ssh_error(mock_run, tmp_path):
    mock_run.return_value = Mock(returncode=255)
    subdir = tmp_path / "myproject"
    subdir.mkdir()

    runner = CliRunner()
    with cwd(subdir):
        result = runner.invoke(entrypoints.remote_init, ["host:path"])

    assert result.exit_code == 1
    assert (
        result.output
        == """\
Failed to create path on remote host host
Please check if host is accessible via SSH
"""
    )

    assert not (subdir / CONFIG_FILE_NAME).exists()
    assert not (subdir / IGNORE_FILE_NAME).exists()


def test_remote_init_fails_if_workspace_is_already_initated(tmp_workspace):
    runner = CliRunner()
    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote_init, ["host2:path2"])

    assert result.exit_code == 1
    assert (
        result.output
        == """\
A configured workspace already exists in the current directory.
If you want to add a new host to it, please use remote-add.
"""
    )

    assert (tmp_workspace / CONFIG_FILE_NAME).exists()
    assert (tmp_workspace / CONFIG_FILE_NAME).read_text() == f"{TEST_CONFIG}\n"


def test_remote_init_fails_on_input_validation(tmp_path):
    runner = CliRunner()

    with cwd(tmp_path):
        result = runner.invoke(entrypoints.remote_init, ["host:path:path"])

    assert result.exit_code == 2
    assert not (tmp_path / CONFIG_FILE_NAME).exists()


def test_remote_commands_fail_on_no_workspace(tmp_path):
    runner = CliRunner()

    results = []
    with cwd(tmp_path):
        results.append(runner.invoke(entrypoints.remote_add, ["host:path"]))
        results.append(runner.invoke(entrypoints.remote_ignore, ["*"]))
        results.append(runner.invoke(entrypoints.remote_host))
        results.append(runner.invoke(entrypoints.remote_set, ["1"]))
        results.append(runner.invoke(entrypoints.remote_pull))
        results.append(runner.invoke(entrypoints.remote_push))
        results.append(runner.invoke(entrypoints.remote_quick, ["echo test"]))
        results.append(runner.invoke(entrypoints.remote, ["echo test"]))
        results.append(runner.invoke(entrypoints.remote_delete))

    for result in results:
        assert result.exit_code == 1
        assert result.output == f"Cannot resolve the remote workspace in {tmp_path}\n"


def test_remote_add_fails_on_input_validation(tmp_path):
    runner = CliRunner()

    with cwd(tmp_path):
        result = runner.invoke(entrypoints.remote_add, ["host:path:path"])

    assert result.exit_code == 2


@patch("remote.util.subprocess.run")
def test_remote_add_adds_host(mock_run, tmp_workspace):
    mock_run.return_value = Mock(returncode=0)
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote_add, ["host:directory"])

    assert result.exit_code == 0
    assert (tmp_workspace / CONFIG_FILE_NAME).exists()
    assert (tmp_workspace / CONFIG_FILE_NAME).read_text() == f"{TEST_CONFIG}\nhost:directory\n"

    mock_run.assert_called_once_with(
        ["ssh", "-tKq", "host", "mkdir -p directory"], stdin=ANY, stdout=ANY, stderr=ANY,
    )


@patch("remote.util.subprocess.run")
def test_remote_add_avoids_duplicates(mock_run, tmp_workspace):
    mock_run.return_value = Mock(returncode=0)
    runner = CliRunner()

    results = []
    with cwd(tmp_workspace):
        results.append(runner.invoke(entrypoints.remote_add, ["host:directory"]))
        results.append(runner.invoke(entrypoints.remote_add, [TEST_CONFIG]))
        results.append(runner.invoke(entrypoints.remote_add, ["host:directory"]))

    for result in results:
        assert result.exit_code == 0
    assert (tmp_workspace / CONFIG_FILE_NAME).exists()
    assert (tmp_workspace / CONFIG_FILE_NAME).read_text() == f"{TEST_CONFIG}\nhost:directory\n"


@patch("remote.util.subprocess.run")
def test_remote_add_fails_on_ssh(mock_run, tmp_workspace):
    mock_run.return_value = Mock(returncode=255)
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote_add, ["host:directory"])

    assert result.exit_code == 1
    assert (tmp_workspace / CONFIG_FILE_NAME).exists()
    assert (tmp_workspace / CONFIG_FILE_NAME).read_text() == f"{TEST_CONFIG}\n"


def test_remote_ignore(tmp_workspace):
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote_ignore, ["*pattern", "other-pattern"])
        # also check there is no duplication
        result_two = runner.invoke(entrypoints.remote_ignore, ["new*.txt", "other-pattern"])

    assert result.exit_code == 0
    assert result_two.exit_code == 0
    assert (
        (tmp_workspace / IGNORE_FILE_NAME).read_text()
        == """\
pull:
push:
both:
*pattern
.remote
.remoteignore
.remoteindex
new*.txt
other-pattern
"""
    )


def test_remote_host(tmp_workspace):
    runner = CliRunner()
    (tmp_workspace / CONFIG_FILE_NAME).write_text(f"{TEST_CONFIG}\nhost:directory\n")

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote_host)

        # Check that result changes if we change host
        runner.invoke(entrypoints.remote_set, ["2"])

    assert result.exit_code == 0
    assert result.output == f"{TEST_HOST}\n"


def test_remote_set(tmp_workspace):
    runner = CliRunner()
    (tmp_workspace / CONFIG_FILE_NAME).write_text(f"{TEST_CONFIG}\nnew-host:directory\n")

    with cwd(tmp_workspace):
        # Check that result changes if we change host
        set_result = runner.invoke(entrypoints.remote_set, ["2"])
        host_result = runner.invoke(entrypoints.remote_host)

        bad_attempt_result = runner.invoke(entrypoints.remote_set, ["10"])

    assert set_result.exit_code == 0
    assert host_result.output == f"new-host\n"
    assert bad_attempt_result.exit_code == 1
    assert (tmp_workspace / INDEX_FILE_NAME).read_text() == "2\n"


@patch("remote.util.subprocess.run")
def test_remote(mock_run, tmp_workspace):
    mock_run.return_value = Mock(returncode=0)
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote, ["echo test >> .file"])

    assert result.exit_code == 0
    assert mock_run.call_count == 3
    mock_run.assert_has_calls(
        [
            call(
                [
                    "rsync",
                    "-rlpmchz",
                    "--copy-unsafe-links",
                    "-e",
                    "ssh -q",
                    "--force",
                    "--exclude-from",
                    ANY,
                    f"{tmp_workspace}/",
                    f"{TEST_HOST}:{TEST_DIR}",
                ],
                stdout=ANY,
                stderr=ANY,
            ),
            call(
                [
                    "ssh",
                    "-tKq",
                    TEST_HOST,
                    """if [ -f .remotes/myproject/.remoteenv ]; then
  source .remotes/myproject/.remoteenv 2>/dev/null 1>/dev/null
fi
cd .remotes/myproject
echo test >> .file
""",
                ],
                stdout=ANY,
                stdin=ANY,
                stderr=ANY,
            ),
            call(
                [
                    "rsync",
                    "-rlpmchz",
                    "--copy-unsafe-links",
                    "-e",
                    "ssh -q",
                    "--force",
                    "--exclude-from",
                    ANY,
                    f"{TEST_HOST}:{TEST_DIR}/",
                    f"{tmp_workspace}",
                ],
                stdout=ANY,
                stderr=ANY,
            ),
        ]
    )


@patch("remote.util.subprocess.run")
def test_remote_execution_fail(mock_run, tmp_workspace):
    mock_run.side_effect = [Mock(returncode=0), Mock(returncode=123), Mock(returncode=0)]
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote, ["echo", "test >> .file"])

    assert result.exit_code == 123
    assert mock_run.call_count == 3
    mock_run.assert_has_calls(
        [
            call(
                [
                    "rsync",
                    "-rlpmchz",
                    "--copy-unsafe-links",
                    "-e",
                    "ssh -q",
                    "--force",
                    "--exclude-from",
                    ANY,
                    f"{tmp_workspace}/",
                    f"{TEST_HOST}:{TEST_DIR}",
                ],
                stdout=ANY,
                stderr=ANY,
            ),
            call(
                [
                    "ssh",
                    "-tKq",
                    TEST_HOST,
                    """if [ -f .remotes/myproject/.remoteenv ]; then
  source .remotes/myproject/.remoteenv 2>/dev/null 1>/dev/null
fi
cd .remotes/myproject
echo 'test >> .file'
""",
                ],
                stdout=ANY,
                stdin=ANY,
                stderr=ANY,
            ),
            call(
                [
                    "rsync",
                    "-rlpmchz",
                    "--copy-unsafe-links",
                    "-e",
                    "ssh -q",
                    "--force",
                    "--exclude-from",
                    ANY,
                    f"{TEST_HOST}:{TEST_DIR}/",
                    f"{tmp_workspace}",
                ],
                stdout=ANY,
                stderr=ANY,
            ),
        ]
    )


@patch("remote.util.subprocess.run")
def test_remote_sync_fail(mock_run, tmp_workspace):
    # first sync fail -> nothing was executed
    mock_run.return_value = Mock(returncode=255)
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote, ["echo test >> .file"])

    assert result.exit_code == 1
    mock_run.assert_called_once_with(
        [
            "rsync",
            "-rlpmchz",
            "--copy-unsafe-links",
            "-e",
            "ssh -q",
            "--force",
            "--exclude-from",
            ANY,
            f"{tmp_workspace}/",
            f"{TEST_HOST}:{TEST_DIR}",
        ],
        stdout=ANY,
        stderr=ANY,
    )


@patch("remote.util.subprocess.run")
def test_remote_quick(mock_run, tmp_workspace):
    mock_run.return_value = Mock(returncode=0)
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote_quick, ["echo", "test"])

    assert result.exit_code == 0
    mock_run.assert_called_once_with(
        [
            "ssh",
            "-tKq",
            TEST_HOST,
            """if [ -f .remotes/myproject/.remoteenv ]; then
  source .remotes/myproject/.remoteenv 2>/dev/null 1>/dev/null
fi
cd .remotes/myproject
echo test
""",
        ],
        stdout=ANY,
        stdin=ANY,
        stderr=ANY,
    )


@patch("remote.util.subprocess.run")
def test_remote_quick_execution_fail(mock_run, tmp_workspace):
    mock_run.return_value = Mock(returncode=15)
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote_quick, ["echo", "test"])

    assert result.exit_code == 15
    mock_run.assert_called_once_with(
        [
            "ssh",
            "-tKq",
            TEST_HOST,
            """if [ -f .remotes/myproject/.remoteenv ]; then
  source .remotes/myproject/.remoteenv 2>/dev/null 1>/dev/null
fi
cd .remotes/myproject
echo test
""",
        ],
        stdout=ANY,
        stdin=ANY,
        stderr=ANY,
    )


@patch("remote.util.subprocess.run")
def test_remote_push(mock_run, tmp_workspace):
    mock_run.return_value = Mock(returncode=0)
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote_push)

    assert result.exit_code == 0
    mock_run.assert_called_once_with(
        [
            "rsync",
            "-rlpmchz",
            "--copy-unsafe-links",
            "-e",
            "ssh -q",
            "--force",
            "-i",
            "--exclude-from",
            ANY,
            f"{tmp_workspace}/",
            f"{TEST_HOST}:{TEST_DIR}",
        ],
        stdout=ANY,
        stderr=ANY,
    )


@patch("remote.util.subprocess.run")
def test_remote_push_mass(mock_run, tmp_workspace):
    (tmp_workspace / CONFIG_FILE_NAME).write_text(f"{TEST_CONFIG}\nnew-host:other-directory\n")

    mock_run.return_value = Mock(returncode=0)
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote_push, "--mass")

    assert result.exit_code == 0
    assert mock_run.call_count == 2
    mock_run.assert_has_calls(
        [
            call(
                [
                    "rsync",
                    "-rlpmchz",
                    "--copy-unsafe-links",
                    "-e",
                    "ssh -q",
                    "--force",
                    "-i",
                    "--exclude-from",
                    ANY,
                    f"{tmp_workspace}/",
                    f"{TEST_HOST}:{TEST_DIR}",
                ],
                stdout=ANY,
                stderr=ANY,
            ),
            call(
                [
                    "rsync",
                    "-rlpmchz",
                    "--copy-unsafe-links",
                    "-e",
                    "ssh -q",
                    "--force",
                    "-i",
                    "--exclude-from",
                    ANY,
                    f"{tmp_workspace}/",
                    f"new-host:other-directory",
                ],
                stdout=ANY,
                stderr=ANY,
            ),
        ]
    )


@patch("remote.util.subprocess.run")
def test_remote_pull(mock_run, tmp_workspace):
    mock_run.return_value = Mock(returncode=0)
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote_pull)

    assert result.exit_code == 0
    mock_run.assert_called_once_with(
        [
            "rsync",
            "-rlpmchz",
            "--copy-unsafe-links",
            "-e",
            "ssh -q",
            "--force",
            "-i",
            "--exclude-from",
            ANY,
            f"{TEST_HOST}:{TEST_DIR}/",
            str(tmp_workspace),
        ],
        stdout=ANY,
        stderr=ANY,
    )


@patch("remote.util.subprocess.run")
def test_remote_pull_subdirs(mock_run, tmp_workspace):
    mock_run.return_value = Mock(returncode=0)
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote_pull, ["build", "dist"])

    assert result.exit_code == 0
    assert mock_run.call_count == 2
    mock_run.assert_has_calls(
        [
            call(
                [
                    "rsync",
                    "-rlpmchz",
                    "--copy-unsafe-links",
                    "-e",
                    "ssh -q",
                    "--force",
                    "-i",
                    f"{TEST_HOST}:{TEST_DIR}/build",
                    f"{tmp_workspace}/build",
                ],
                stdout=ANY,
                stderr=ANY,
            ),
            call(
                [
                    "rsync",
                    "-rlpmchz",
                    "--copy-unsafe-links",
                    "-e",
                    "ssh -q",
                    "--force",
                    "-i",
                    f"{TEST_HOST}:{TEST_DIR}/dist",
                    f"{tmp_workspace}/dist",
                ],
                stdout=ANY,
                stderr=ANY,
            ),
        ]
    )


@patch("remote.util.subprocess.run")
def test_remote_delete(mock_run, tmp_workspace):
    mock_run.return_value = Mock(returncode=0)
    runner = CliRunner()

    with cwd(tmp_workspace):
        result = runner.invoke(entrypoints.remote_delete)

    assert result.exit_code == 0
    mock_run.assert_called_once_with(
        ["ssh", "-tKq", TEST_HOST, f"rm -rf {TEST_DIR}"], stdin=ANY, stdout=ANY, stderr=ANY,
    )



from pathlib import Path
from typing import List, Tuple

from remote.exceptions import ConfigurationError

from . import ConfigurationMedium, WorkspaceConfig
from .classic import ClassicConfigurationMedium

CONFIG_MEDIUMS: List[ConfigurationMedium] = [ClassicConfigurationMedium()]


def resolve_workspace_root(working_dir: Path) -> Tuple[ConfigurationMedium, Path]:
    """Find and return the directory in this tree that has remote-ing set up"""
    possible_directory = working_dir
    root = Path("/")
    while possible_directory != root:
        for meduim in CONFIG_MEDIUMS:
            if meduim.is_workspace_root(possible_directory):
                return meduim, possible_directory
        possible_directory = possible_directory.parent

    raise ConfigurationError(f"Cannot resolve the remote workspace in {working_dir}")


def load_cwd_workspace_config() -> WorkspaceConfig:
    """Determine current working directory's workspace config type and load it

    Supports only classical config layout now, but will be extended once the new type is added
    """
    working_dir = Path.cwd()
    medium, root = resolve_workspace_root(working_dir)
    return medium.load_config(root)


def get_configuration_medium(config: WorkspaceConfig) -> ConfigurationMedium:
    """We have only one medium for now"""
    for medium in CONFIG_MEDIUMS:
        if medium.is_workspace_root(config.root):
            return medium

    # If there is no medium found, the config is newly created, so we return a default one
    return ClassicConfigurationMedium()


def save_config(config: WorkspaceConfig):
    """Save config using the proper medium"""
    config_medium = get_configuration_medium(config)
    config_medium.save_config(config)



import logging
import re
import sys

from functools import wraps
from pathlib import Path
from typing import List

import click

from .configuration import WorkspaceConfig
from .configuration.discovery import get_configuration_medium, load_cwd_workspace_config, save_config
from .exceptions import RemoteError
from .workspace import SyncedWorkspace

BASE_LOGGING_FORMAT = "%(message)s"
CONNECTION_STRING_FORMAT_REGEX = re.compile(r"^[-\w]+(\.[-\w]+)*(:(/)?[-.\w\s]+(/[-.\w\s]+)*)?(/)?$")
DEFAULT_CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
EXECUTION_CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"], ignore_unknown_options=True, allow_interspersed_args=False
)


def log_exceptions(f):
    """A decorator that prints the custom exceptions and exit, but propagates internal ones"""

    @wraps(f)
    def wrapper(*args, **kwards):
        try:
            f(*args, **kwards)
        except Exception as e:
            if isinstance(e, RemoteError):
                click.secho(str(e), fg="yellow")
                sys.exit(1)
            raise

    return wrapper


def validate_connection_string(ctx, param, value):
    matcher = CONNECTION_STRING_FORMAT_REGEX.match(value)
    if matcher is None:
        raise click.BadParameter(
            "Please fix value to match the specified format for connection string", ctx=ctx, param=param
        )
    return value


def _add_remote_host(config: WorkspaceConfig, connection: str):
    """Add a new remote host to the workspace config, check the connection, and save it if connection is ok

    :param config: the workspace config decription object
    :param connection: connection string in format of 'host-name[:remote_dir]'
    """
    parts = connection.split(":")
    remote_host = parts[0]
    config_medium = get_configuration_medium(config)
    remote_dir = config_medium.generate_remote_directory(config) if len(parts) == 1 else Path(parts[1])

    added, index = config.add_remote_host(remote_host, remote_dir)
    if not added:
        click.echo(f"{connection} already exists in config")
        sys.exit(0)

    # Check if we can connect to the remote host and create a directory there
    workspace = SyncedWorkspace.from_config(config, config.root, index)
    code = workspace.execute(f"mkdir -p {workspace.remote.directory}", simple=True, raise_on_error=False)
    if code != 0:
        click.secho(f"Failed to create {workspace.remote.directory} on remote host {remote_host}", fg="yellow")
        click.secho(f"Please check if host is accessible via SSH", fg="yellow")
        sys.exit(1)

    click.echo(f"Created remote directory at {workspace.remote.host}:{workspace.remote.directory}")
    click.echo(f"Remote is configured and ready to use")

    # No errors when executing the above code means we can save the config
    config_medium.save_config(config)


@click.command(context_settings=DEFAULT_CONTEXT_SETTINGS)
@click.argument("connection", metavar="host-name[:remote_dir]", callback=validate_connection_string)
@log_exceptions
def remote_add(connection: str):
    """Add one more host for remote connection to a config file"""

    config = load_cwd_workspace_config()
    _add_remote_host(config, connection)


@click.command(context_settings=DEFAULT_CONTEXT_SETTINGS)
@click.argument("connection", metavar="host-name[:remote_dir]", callback=validate_connection_string)
@log_exceptions
def remote_init(connection: str):
    """Initiate workspace for the remote execution in the current working directory"""

    try:
        load_cwd_workspace_config()
        click.secho("A configured workspace already exists in the current directory.", fg="yellow")
        click.secho("If you want to add a new host to it, please use remote-add.", fg="yellow")
        sys.exit(1)
    except RemoteError:
        # we expect it to fail. It means we don't overwrite an existing workspace
        pass

    config = WorkspaceConfig.empty(Path.cwd())
    _add_remote_host(config, connection)

    # help out with .gitignore if we are in a git repository
    if not (config.root / ".git").exists():
        return

    # make sure we don't keep adding to .gitignore
    gitignore = config.root / ".gitignore"
    if gitignore.exists():
        for line in gitignore.read_text().splitlines():
            if line.startswith(".remote"):
                return

    with gitignore.open("a") as f:
        f.write("\n")
        f.write(".remote*")
        f.write("\n")

    click.echo("Added '.remote*' to .gitignore")


@click.command(context_settings=DEFAULT_CONTEXT_SETTINGS)
@click.argument("ignore", nargs=-1, required=True)
@log_exceptions
def remote_ignore(ignore: List[str]):
    """Add new IGNORE patterns to the ignores list

    IGNORE pattern should be a string in rsync-friendly format.
    """

    config = load_cwd_workspace_config()
    config.ignores.add_ignores(ignore)
    save_config(config)


@click.command(context_settings=DEFAULT_CONTEXT_SETTINGS)
@log_exceptions
def remote_host():
    """Print the default remote host in use and exit"""
    workspace = SyncedWorkspace.from_cwd()
    click.echo(workspace.remote.host)


@click.command(context_settings=DEFAULT_CONTEXT_SETTINGS)
@click.argument("index", type=int)
@log_exceptions
def remote_set(index: int):
    """Set a new default remote host for the workspace

    INDEX is an index of host in config file to use by default (strating from 1)
    """

    config = load_cwd_workspace_config()
    if len(config.configurations) < index:
        click.secho(
            f"Index is too big ({index}). Only have {len(config.configurations)} hosts to choose from.", fg="yellow"
        )
        sys.exit(1)
    elif index < 1:
        click.secho(f"Index should be 1 or higher", fg="yellow")
        sys.exit(1)
    # we use 0-base index internally
    index = index - 1
    config.default_configuration = index
    save_config(config)

    click.echo(f"Remote host is set to {config.configurations[index].host}")


@click.command(context_settings=EXECUTION_CONTEXT_SETTINGS)
@click.option("-n", "--dry-run", is_flag=True, help="do a dry run of the whole cycle")
@click.option("-m", "--mirror", is_flag=True, help="mirror local files on the remote host")
@click.option("-v", "--verbose", is_flag=True, help="increase verbosity")
@click.option("-e", is_flag=True, help="(deprecated) kept for backward compatibility, noop")
@click.argument("command", nargs=-1, required=True)
@log_exceptions
def remote(command: List[str], dry_run: bool, mirror: bool, verbose: bool, e: bool):
    """Sync local workspace files to remote machine, execute the COMMAND and sync files back regardless of the result"""

    if verbose:
        logging.basicConfig(level=logging.INFO, format=BASE_LOGGING_FORMAT)

    workspace = SyncedWorkspace.from_cwd()
    exit_code = workspace.execute_in_synced_env(command, dry_run=dry_run, verbose=verbose, mirror=mirror)
    if exit_code != 0:
        click.secho(f"Remote command exited with {exit_code}", fg="yellow")

    sys.exit(exit_code)


@click.command(context_settings=EXECUTION_CONTEXT_SETTINGS)
@click.argument("command", nargs=-1, required=True)
@log_exceptions
def remote_quick(command: List[str]):
    """Execute the COMMAND remotely"""
    workspace = SyncedWorkspace.from_cwd()
    code = workspace.execute(command, raise_on_error=False)
    sys.exit(code)


@click.command(context_settings=DEFAULT_CONTEXT_SETTINGS)
@click.option("-n", "--dry-run", is_flag=True, help="do a dry run of a pull")
@click.option("-v", "--verbose", is_flag=True, help="increase verbosity")
@click.argument("path", nargs=-1)
@log_exceptions
def remote_pull(dry_run: bool, verbose: bool, path: List[str]):
    """Bring in files from the default remote directory to local workspace.
    Optionally bring in PATH instead of the whole workspace.

    PATH is a path of file or directory to bring back relative to the remote workspace root.
    All sync exclude rules will be omitted if PATH is provided.
    """

    if verbose:
        logging.basicConfig(level=logging.INFO, format=BASE_LOGGING_FORMAT)

    workspace = SyncedWorkspace.from_cwd()
    if not path:
        workspace.pull(info=True, verbose=verbose, dry_run=dry_run)
        return

    for subpath in path:
        workspace.pull(info=True, verbose=verbose, dry_run=dry_run, subpath=subpath)


@click.command(context_settings=DEFAULT_CONTEXT_SETTINGS)
@click.option("-n", "--dry-run", is_flag=True, help="do a dry run of a push")
@click.option("-m", "--mirror", is_flag=True, help="mirror local files on the remote host")
@click.option("-v", "--verbose", is_flag=True, help="increase verbosity")
@click.option(
    "--mass", is_flag=True, help="push files to all available remote workspaces instead of pushing to the default one"
)
@log_exceptions
def remote_push(dry_run: bool, mirror: bool, verbose: bool, mass: bool):
    """Push local workspace files to the remote directory"""

    if verbose:
        logging.basicConfig(level=logging.INFO, format=BASE_LOGGING_FORMAT)

    workspaces = SyncedWorkspace.from_cwd_mass() if mass else [SyncedWorkspace.from_cwd()]
    for workspace in workspaces:
        workspace.push(info=True, verbose=verbose, dry_run=dry_run, mirror=mirror)


@click.command(context_settings=DEFAULT_CONTEXT_SETTINGS)
@log_exceptions
def remote_delete():
    """Delete the remote directory"""
    workspace = SyncedWorkspace.from_cwd()
    workspace.clear_remote()
    click.echo(f"Successfully deleted {workspace.remote.directory} on host {workspace.remote.host}")


@click.command(context_settings=DEFAULT_CONTEXT_SETTINGS)
@log_exceptions
def remote_explain():
    click.secho("Sorry, remote-explain is not yet implemented in the new version of Remote.", fg="yellow")
    click.secho("Please use the old one if you need it", fg="yellow")
    sys.exit(1)


@click.command(context_settings=DEFAULT_CONTEXT_SETTINGS)
@log_exceptions
def mremote():
    click.secho("Sorry, mremote is not yet implemented in the new version of Remote.", fg="yellow")
    click.secho("Please use the old one if you need it", fg="yellow")
    sys.exit(1)


@click.command(context_settings=DEFAULT_CONTEXT_SETTINGS)
@log_exceptions
def mremote_push():
    click.secho("mremote-push is deprecated. Please use 'remote-push --mass' instead.", fg="yellow")



class RemoteError(Exception):
    """A base class for all remote's custom exception"""


class RemoteConnectionError(RemoteError):
    """Remote wasn't able to connect to remote host"""


class RemoteExecutionError(RemoteError):
    """A command executed remotely exited with non-zero status"""


class ConfigurationError(RemoteError):
    """The workspace configuration is incorrect"""



from pathlib import Path
from unittest.mock import ANY, MagicMock, call, patch

import pytest

from remote.configuration import RemoteConfig
from remote.workspace import SyncedWorkspace


@pytest.fixture
def workspace(workspace_config):
    workspace_config.ignores.pull.append("build")
    working_dir = workspace_config.root / "foo" / "bar"
    working_dir.mkdir(parents=True)
    return SyncedWorkspace.from_config(workspace_config, working_dir)


def test_create_workspace(workspace_config):
    working_dir = workspace_config.root / "foo" / "bar"
    workspace = SyncedWorkspace.from_config(workspace_config, working_dir)

    assert workspace.local_root == workspace_config.root
    assert workspace.remote == workspace_config.configurations[0]
    assert workspace.remote_working_dir == workspace_config.configurations[0].directory / "foo" / "bar"
    assert workspace.ignores == workspace_config.ignores


def test_create_workspace_selects_proper_remote_host(workspace_config):
    working_dir = workspace_config.root / "foo" / "bar"
    workspace_config.configurations.append(
        RemoteConfig(
            host="other-host.example.com", directory=Path("other/dir"), shell="bash", shell_options="some options"
        )
    )
    workspace_config.default_configuration = 1

    # workspace should select host from workspace_config.default_configuration
    workspace = SyncedWorkspace.from_config(workspace_config, working_dir)
    assert workspace.local_root == workspace_config.root
    assert workspace.remote == workspace_config.configurations[1]
    assert workspace.remote_working_dir == workspace_config.configurations[1].directory / "foo" / "bar"
    assert workspace.ignores == workspace_config.ignores

    # now it should select host from overrid
    workspace = SyncedWorkspace.from_config(workspace_config, working_dir, config_num=0)
    assert workspace.local_root == workspace_config.root
    assert workspace.remote == workspace_config.configurations[0]
    assert workspace.remote_working_dir == workspace_config.configurations[0].directory / "foo" / "bar"
    assert workspace.ignores == workspace_config.ignores


@patch("remote.util.subprocess.run")
def test_clear_remote_workspace(mock_run, workspace):
    mock_run.return_value = MagicMock(returncode=0)

    workspace.clear_remote()

    # clear should always delete remote root regardless of what the workign dir is
    mock_run.assert_called_once_with(
        ["ssh", "-tKq", workspace.remote.host, f"rm -rf {workspace.remote.directory}"],
        stderr=ANY,
        stdin=ANY,
        stdout=ANY,
    )


@patch("remote.util.subprocess.run")
def test_push(mock_run, workspace):
    mock_run.return_value = MagicMock(returncode=0)

    workspace.push()
    mock_run.assert_called_once_with(
        [
            "rsync",
            "-rlpmchz",
            "--copy-unsafe-links",
            "-e",
            "ssh -q",
            "--force",
            f"{workspace.local_root}/",
            f"{workspace.remote.host}:{workspace.remote.directory}",
        ],
        stderr=ANY,
        stdout=ANY,
    )


@patch("remote.util.subprocess.run")
def test_pull(mock_run, workspace):
    mock_run.return_value = MagicMock(returncode=0)

    workspace.pull()
    mock_run.assert_called_once_with(
        [
            "rsync",
            "-rlpmchz",
            "--copy-unsafe-links",
            "-e",
            "ssh -q",
            "--force",
            "--exclude-from",
            ANY,
            f"{workspace.remote.host}:{workspace.remote.directory}/",
            f"{workspace.local_root}",
        ],
        stderr=ANY,
        stdout=ANY,
    )


@patch("remote.util.subprocess.run")
def test_pull_with_subdir(mock_run, workspace):
    mock_run.return_value = MagicMock(returncode=0)

    workspace.pull(subpath="some-path")
    mock_run.assert_called_once_with(
        [
            "rsync",
            "-rlpmchz",
            "--copy-unsafe-links",
            "-e",
            "ssh -q",
            "--force",
            f"{workspace.remote.host}:{workspace.remote.directory}/some-path",
            f"{workspace.local_root}/some-path",
        ],
        stderr=ANY,
        stdout=ANY,
    )


@patch("remote.util.subprocess.run")
def test_execute(mock_run, workspace):
    mock_run.return_value = MagicMock(returncode=0)

    code = workspace.execute(["echo", "Hello World!"])
    mock_run.assert_called_once_with(
        [
            "ssh",
            "-tKq",
            workspace.remote.host,
            """\
if [ -f remote/dir/.remoteenv ]; then
  source remote/dir/.remoteenv 2>/dev/null 1>/dev/null
fi
cd remote/dir/foo/bar
echo 'Hello World!'
""",
        ],
        stderr=ANY,
        stdin=ANY,
        stdout=ANY,
    )
    assert code == 0


@patch("remote.util.subprocess.run")
def test_execute_and_sync(mock_run, workspace):
    mock_run.side_effect = [MagicMock(returncode=0), MagicMock(returncode=10), MagicMock(returncode=0)]

    code = workspace.execute_in_synced_env(["echo", "Hello World!"])
    mock_run.assert_has_calls(
        [
            call(
                [
                    "rsync",
                    "-rlpmchz",
                    "--copy-unsafe-links",
                    "-e",
                    "ssh -q",
                    "--force",
                    f"{workspace.local_root}/",
                    f"{workspace.remote.host}:{workspace.remote.directory}",
                ],
                stderr=ANY,
                stdout=ANY,
            ),
            call(
                [
                    "ssh",
                    "-tKq",
                    workspace.remote.host,
                    """\
if [ -f remote/dir/.remoteenv ]; then
  source remote/dir/.remoteenv 2>/dev/null 1>/dev/null
fi
cd remote/dir/foo/bar
echo 'Hello World!'
""",
                ],
                stderr=ANY,
                stdin=ANY,
                stdout=ANY,
            ),
            call(
                [
                    "rsync",
                    "-rlpmchz",
                    "--copy-unsafe-links",
                    "-e",
                    "ssh -q",
                    "--force",
                    "--exclude-from",
                    ANY,
                    f"{workspace.remote.host}:{workspace.remote.directory}/",
                    f"{workspace.local_root}",
                ],
                stderr=ANY,
                stdout=ANY,
            ),
        ]
    )
    assert code == 10



import logging

from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union

from .configuration import RemoteConfig, SyncIgnores, WorkspaceConfig
from .configuration.discovery import load_cwd_workspace_config
from .util import prepare_shell_command, rsync, ssh

logger = logging.getLogger(__name__)


@dataclass
class SyncedWorkspace:
    """A configured remote execution workspace"""

    # absolute path to the root of the local project
    local_root: Path
    # information about the remote host to use
    remote: RemoteConfig
    # remote directory to use as working directory, relative to users home directory
    remote_working_dir: Path
    # sync ignore file patterns
    ignores: SyncIgnores

    @classmethod
    def from_config(
        cls, config: WorkspaceConfig, working_dir: Path, config_num: Optional[int] = None
    ) -> "SyncedWorkspace":
        """Create a workspace from configuration object

        :param config: workspace config
        :param working_dir: a working directory inside the workspace config
        :param config_num: if present, overrides the default remote host to use
        """
        working_dir = working_dir.relative_to(config.root)
        if config_num is None:
            config_num = config.default_configuration
        remote_config = config.configurations[config_num]
        remote_working_dir = remote_config.directory / working_dir

        return cls(
            local_root=config.root, remote=remote_config, remote_working_dir=remote_working_dir, ignores=config.ignores,
        )

    @classmethod
    def from_cwd(cls) -> "SyncedWorkspace":
        """Load a workspace from current working directory of user"""
        config = load_cwd_workspace_config()
        return cls.from_config(config, Path.cwd())

    @classmethod
    def from_cwd_mass(cls) -> List["SyncedWorkspace"]:
        """Load all possible workspaces from current working directory of user"""
        config = load_cwd_workspace_config()

        workspaces = []
        for i in range(len(config.configurations)):
            workspaces.append(cls.from_config(config, Path.cwd(), i))

        return workspaces

    def _generate_command(self, command: str) -> str:
        return f"""\
if [ -f {self.remote.directory}/.remoteenv ]; then
  source {self.remote.directory}/.remoteenv 2>/dev/null 1>/dev/null
fi
cd {self.remote_working_dir}
{command}
"""

    def execute_in_synced_env(
        self, command: Union[str, List[str]], simple=False, verbose=False, dry_run=False, mirror=False
    ) -> int:
        """Execute a command remotely using ssh. Push the local files to remote location before that and
        pull them back after command was executed regardless of the result.

        This command won't throw an exception if remote process fails

        :param command: a command to be executed or its parts
        :param simple: True if command don't need to be preformatted and wrapped before execution.
                       commands with simple will be executed from user's remote home directory
        :param dry_run: don't sync files. log the command to be executed but don't run it
        :param verbose: use verbose logging when running rsync and remote execution
        :param mirror: mirror local files remotely. It will remove ALL the remote files in the directory
                       that weren't synced from local workspace

        :returns: an exit code of a remote process
        """

        self.push(dry_run=dry_run, verbose=verbose, mirror=mirror)
        exit_code = self.execute(command, simple=simple, dry_run=dry_run, raise_on_error=False)
        if exit_code != 0:
            logger.info(f"Remote command exited with {exit_code}")
        self.pull(dry_run=dry_run, verbose=verbose)
        return exit_code

    def execute(self, command: Union[str, List[str]], simple=False, dry_run=False, raise_on_error=True) -> int:
        """Execute a command remotely using ssh

        :param command: a command to be executed or its parts
        :param simple: True if command don't need to be preformatted and wrapped before execution.
                       commands with simple will be executed from user's remote home directory
        :param dry_run: log the command to be executed but don't run it
        :param raise_on_error: raise exception if error code was other than 0

        :returns: an exit code of a remote process
        """
        formatted_command = prepare_shell_command(command)
        if not simple:
            formatted_command = self._generate_command(formatted_command)

        return ssh(self.remote.host, formatted_command, dry_run=dry_run, raise_on_error=raise_on_error)

    def push(self, info=False, verbose=False, dry_run=False, mirror=False):
        """Push local workspace files to remote directory

        :param info: use info logging when running rsync
        :param verbose: use verbose logging when running rsync
        :param dry_run: use dry_run parameter when running rsync
        :param mirror: mirror local files remotely. It will remove ALL the remote files in the directory
                       that weren't synced from local workspace
        """
        src = f"{self.local_root}/"
        dst = f"{self.remote.host}:{self.remote.directory}"
        ignores = self.ignores.compile_push_ignores()
        rsync(src, dst, info=info, verbose=verbose, dry_run=dry_run, mirror=mirror, excludes=ignores)

    def pull(self, info=False, verbose=False, dry_run=False, subpath=None):
        """Pull remote files to local workspace

        :param info: use info logging when running rsync
        :param verbose: use verbose logging when running rsync
        :param dry_run: use dry_run parameter when running rsync
        :param subpath: a specific path to bring in. If provided, subpath will be synced
                        even if it is ignored by workspace rules
        """
        if subpath is not None:
            src = f"{self.remote.host}:{self.remote.directory}/{subpath}"
            dst = str(self.local_root / subpath)
            rsync(src, dst, info=info, verbose=verbose, dry_run=dry_run)
            return

        src = f"{self.remote.host}:{self.remote.directory}/"
        dst = str(self.local_root)
        ignores = self.ignores.compile_pull_ignores()
        rsync(src, dst, info=info, verbose=verbose, dry_run=dry_run, excludes=ignores)

    def clear_remote(self):
        """Remove remote directory"""
        self.execute(f"rm -rf {self.remote.directory}", simple=True)



from unittest.mock import ANY, MagicMock, patch

import pytest

from remote.exceptions import RemoteConnectionError, RemoteExecutionError
from remote.util import _temp_file, prepare_shell_command, rsync, ssh


def test_temp_file():
    file = _temp_file(["1", "", "2", "3"])
    assert file.exists()
    assert (
        file.read_text()
        == """\
1

2
3
"""
    )
    assert file.name.startswith("remote.")


def test_rsync_copies_files(tmp_path):
    src = tmp_path / "src"
    src.mkdir()
    (src / "first").write_text("TEST first")
    (src / "second").write_text("TEST second")
    (src / "third.txt").write_text("TEST third")
    (src / "fourth.txt").write_text("TEST fourth")

    dst = tmp_path / "dst"
    rsync(f"{src}/", str(dst), excludes=["f*"], includes=["*.txt"])

    assert dst.exists()
    assert not (dst / "first").exists()
    assert (dst / "second").exists()
    assert (dst / "second").read_text() == "TEST second"
    assert (dst / "third.txt").exists()
    assert (dst / "third.txt").read_text() == "TEST third"
    assert (dst / "fourth.txt").exists()
    assert (dst / "fourth.txt").read_text() == "TEST fourth"


def test_rsync_copies_files_with_mirror(tmp_path):
    src = tmp_path / "src"
    src.mkdir()
    (src / "first").write_text("TEST first")
    (src / "second").write_text("TEST second")
    (src / "third.txt").write_text("TEST third")
    (src / "fourth.txt").write_text("TEST fourth")

    dst = tmp_path / "dst"
    dst.mkdir()
    (dst / "first").write_text("TEST first")

    # since we use mirror=True "first" should be deleted in dst
    rsync(f"{src}/", str(dst), excludes=["f*"], mirror=True)

    assert dst.exists()
    assert not (dst / "first").exists()
    assert (dst / "second").exists()
    assert (dst / "second").read_text() == "TEST second"
    assert (dst / "third.txt").exists()
    assert (dst / "third.txt").read_text() == "TEST third"
    assert not (dst / "fourth.txt").exists()


@patch("remote.util.subprocess.run")
def test_rsync_respects_all_options(mock_run):
    mock_run.return_value = MagicMock(returncode=0)
    rsync(f"src/", "dst", info=True, verbose=True, mirror=True, dry_run=True, extra_args=["--some-extra"])

    mock_run.assert_called_once_with(
        [
            "rsync",
            "-rlpmchz",
            "--copy-unsafe-links",
            "-e",
            "ssh -q",
            "--force",
            "-i",
            "-v",
            "-n",
            "--delete",
            "--delete-after",
            "--delete-excluded",
            "--some-extra",
            "src/",
            "dst",
        ],
        stdout=ANY,
        stderr=ANY,
    )


@patch("remote.util.subprocess.run")
def test_rsync_throws_exception_on_bad_return_code(mock_run):
    mock_run.return_value = MagicMock()
    mock_run.return_value.returncode = 1

    with pytest.raises(RemoteConnectionError):
        rsync(f"src/", "dst")


@pytest.mark.parametrize("returncode", [0, 1])
@patch("remote.util.subprocess.run")
@patch("remote.util._temp_file")
def test_rsync_always_removes_temporary_files(mock_temp_file, mock_run, returncode, tmp_path):
    mock_run.return_value = MagicMock(returncode=returncode)

    files = []

    def ignore_file(_):
        new_path = tmp_path / f"file{len(files)}"
        new_path.write_text("TEST")
        files.append(new_path)
        return new_path

    mock_temp_file.side_effect = ignore_file

    try:
        rsync(f"src/", "dst", excludes=["f*"], includes=["*.txt"])
    except Exception:
        pass

    assert len(files) == 2
    for file in files:
        assert not file.exists()


@patch("remote.util.subprocess.run")
def test_ssh(mock_run):
    mock_run.return_value = MagicMock(returncode=0)

    code = ssh("my-host.example.com", "exit 0")

    assert code == 0
    mock_run.assert_called_once_with(
        ["ssh", "-tKq", "my-host.example.com", "exit 0"], stdout=ANY, stderr=ANY, stdin=ANY
    )


@pytest.mark.parametrize("returncode, error", [(255, RemoteConnectionError), (1, RemoteExecutionError)])
@patch("remote.util.subprocess.run")
def test_ssh_raises_exception(mock_run, returncode, error):
    mock_run.return_value = MagicMock(returncode=returncode)

    with pytest.raises(error):
        ssh("my-host.example.com", f"exit {returncode}")

    mock_run.assert_called_once_with(
        ["ssh", "-tKq", "my-host.example.com", f"exit {returncode}"], stdout=ANY, stderr=ANY, stdin=ANY
    )


@pytest.mark.parametrize("returncode", [255, 1])
@patch("remote.util.subprocess.run")
def test_ssh_returns_error_code_if_configured(mock_run, returncode):
    mock_run.return_value = MagicMock()
    mock_run.return_value.returncode = returncode

    code = ssh("my-host.example.com", f"exit {returncode}", raise_on_error=False)

    assert code == returncode
    mock_run.assert_called_once_with(
        ["ssh", "-tKq", "my-host.example.com", f"exit {returncode}"], stdout=ANY, stderr=ANY, stdin=ANY
    )


@patch("remote.util.subprocess.run")
def test_ssh_no_execute_on_dry_run(mock_run):
    code = ssh("my-host.example.com", "exit 1", dry_run=True)

    assert code == 0
    mock_run.assert_not_called()


@pytest.mark.parametrize(
    "command, expected",
    [
        (["ls", "-l"], "ls -l"),
        (["ls -l"], "ls -l"),
        ("ls -l", "ls -l"),
        (["echo", "some text here"], "echo 'some text here'"),
        (["echo", "some\ntext\nhere"], "echo 'some\ntext\nhere'"),
        (["du", "-sh", "-C", "/some/path/with whitespace/"], "du -sh -C '/some/path/with whitespace/'"),
    ],
)
def test_prepare_shell_command(command, expected):
    result = prepare_shell_command(command)

    assert result == expected



import os
import setuptools


def get_version():
    root = os.path.dirname(__file__)
    changelog = os.path.join(root, "CHANGELOG")
    with open(changelog) as f:
        return f.readline().strip()


setuptools.setup(
    name="remote-exec",
    version=get_version(),
    url="https://github.com/shirshanka/remote",
    author="Shirshanka Das",
    license="BSD-2-CLAUSE",
    description="A CLI to sync codebases and execute commands remotely",
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="src"),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "remote = remote.entrypoints:remote",
            "remote-add = remote.entrypoints:remote_add",
            "remote-delete = remote.entrypoints:remote_delete",
            "remote-explain = remote.entrypoints:remote_explain",
            "remote-host = remote.entrypoints:remote_host",
            "remote-ignore = remote.entrypoints:remote_ignore",
            "remote-init = remote.entrypoints:remote_init",
            "remote-pull = remote.entrypoints:remote_pull",
            "remote-push = remote.entrypoints:remote_push",
            "remote-quick = remote.entrypoints:remote_quick",
            "remote-set = remote.entrypoints:remote_set",
            "mremote = remote.entrypoints:mremote",
            "mremote-push = remote.entrypoints:mremote_push",
        ],
    },
    install_requires=['dataclasses; python_version<="3.6"', "click>=7.1.1"],
)
